--[[
    ManagerFrameLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    ManagerFrameLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ManagerFrameLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ManagerFrameLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local managerFrameLib = LibStub:NewLibrary("ManagerFrameLib-1.0", 1);
if not managerFrameLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
local tM2TimerRendererLib = LibStub("TM2TimerRendererLib-1.0");
local libSinkTimerRendererLib = LibStub("LibSinkTimerRendererLib-1.0");
local actionLib = LibStub("ActionLib-1.0");
local actionBarLib = LibStub("ActionBarLib-1.0");
local timerLib = LibStub("TimerLib-1.0");
-- end import

local function rankToRoman(rank)
  if rank>18 then
    error("Unsupported rank");
    return;
  end
  local m5 = rank%5;
  local before, after = "", "";
  if m5==1 then
    after = "I";  
  elseif m5==2 then
    after = "II";
  elseif m5==3 then
    after = "III";
  elseif m5==4 then
    before = "I";
  end
  local rest = math.floor((rank + 1) / 5);
  if rest == 0 then
    return before..after; -- 0 to 3
  elseif rest == 1 then
    return before.."V"..after; -- 4 to 8
  elseif rest == 2 then
    return before.."X"..after; -- 9 to 13
  elseif rest == 3 then
    return "X"..before.."V"..after;  -- 14 to 18
  end  
end

-- this function is necessary to spell action
local function transformSpellArg(initialSpellArg, action, rank, targetName)
  local res = {};
  if initialSpellArg.totemName and rank~=-1 and rank~=1 then
    res.name = initialSpellArg.totemName.." "..rankToRoman(rank);
  else
    res.name = initialSpellArg.totemName or initialSpellArg.buffName or initialSpellArg.name;
  end
  res.target = targetName;
  if initialSpellArg.duration == nil then
    res.duration = 300; -- default to 5 minutes
  elseif type(initialSpellArg.duration)=="function" then
    res.duration = initialSpellArg.duration(rank);
  else
    res.duration = initialSpellArg.duration;
  end
  res.texture = action:GetTexture();
  if initialSpellArg.tickFrequency ~= nil then
    res.tickFrequency = initialSpellArg.tickFrequency;
    res.tickStart = initialSpellArg.tickStart or 0.5;
  end
  res.updater = initialSpellArg.updater;
  if res.updater ~= nil then
     res.updater.isRaidEffect = initialSpellArg.isRaidEffect;
     res.updater.effect = initialSpellArg.effect;
     res.updater.found = false;
  end
  return res;
end

local function earthShieldTransformSpellArg(initialSpellArg, action, _, targetName)
  local res = {};
  if UnitName("player")==targetName then
    res.duration = 0;
  else
    res.name = initialSpellArg.name;
    res.duration = initialSpellArg.duration;  
    res.target = targetName;
    res.texture = action:GetTexture();
    res.updater = initialSpellArg.updater;
    res.updater.found = false;
    res.updater.targetName = targetName;
    res.updater.targetPath = environmentLib.environment:FindPath(targetName);
  end

  return res;
end

local function transformItemArg(initialSpellArg, action, targetName)
  local res = {};
  res.name = initialSpellArg.totemName or initialSpellArg.name;
  if initialSpellArg.duration == nil then
    res.duration = 120; -- default to 2 minutes
  else
    res.duration = initialSpellArg.duration;
  end
  res.texture = action:GetTexture();
  res.updater = initialSpellArg.updater;
  if res.updater ~= nil then
     res.updater.isRaidEffect = initialSpellArg.isRaidEffect;
     res.updater.effect = initialSpellArg.effect;
     res.updater.found = false;
  end
  return res;
end


--[[
  The ManagerFrame class.
--]]

-- ManagerFrame
local ManagerFrame, ManagerFrameFactory, managerFrameFactory = factoryLib.Factory.CreateType();
managerFrameLib.ManagerFrame = ManagerFrame;
managerFrameLib.ManagerFrameFactory = ManagerFrameFactory;
managerFrameLib.managerFrameFactory = managerFrameFactory;

function ManagerFrame:ToggleLocked()
  if self.locked then
    -- transparent colors
    self.frame:SetBackdropColor(0, 0, 0, 0);
    self.frame:SetBackdropBorderColor(.3, .3, .3, 0);
    self.frame:EnableMouse(false);
    self.locked = false;
  else
    self.frame:SetBackdropColor(0, 0, 0, 1);
    self.frame:SetBackdropBorderColor(.3, .3, .3, 1);
    self.frame:EnableMouse(true);
    self.locked = true;
  end
end

function ManagerFrame:ToggleVisibility()
  if self.visibility then
    self.frame:Hide();
    self.visibility = false;
  else
    self.frame:Show();
    self.visibility = true;  
  end
end

function ManagerFrame:GetVisibility()
  return self.visibility;
end

ManagerFrame.TWISTED_LEFT_DIRECTION = 1;
ManagerFrame.TWISTED_RIGHT_DIRECTION = 2;
ManagerFrame.LEFT_DIRECTION = 3;
ManagerFrame.RIGHT_DIRECTION = 4;

ManagerFrame.actionBarSkin = {verticalGap=30, horizontalGap=30};
ManagerFrame.timerSkin = {visualRenderer = tM2TimerRendererLib.tM2TimerRendererFactory, verticalGap=40, horizontalGap=100};
ManagerFrame.minimumGap = 35;

function ManagerFrame:SetDirection(_, direction)
  local gap = max(self.actionBarSkin.verticalGap, self.timerSkin.verticalGap, self.minimumGap);
  if direction == self.LEFT_DIRECTION or direction == self.TWISTED_LEFT_DIRECTION then
    for z, actionBarRenderer in ipairs(self.actionBarRenderer) do
      actionBarRenderer:SetPosition(self.frame, 5+self.actionBarSkin.horizontalGap/2, 
                                    5+gap*(z-.5));
      if direction == self.TWISTED_LEFT_DIRECTION then
        actionBarRenderer:SetAngle(7*math.pi/8);
      else
        actionBarRenderer:SetAngle(math.pi);
      end
    end
    for z, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetPosition(self.frame, 10+self.actionBarSkin.horizontalGap, 
                                      5+gap*(z-.5));
      visualTimerRenderer:SetDirection(1);  
    end
  elseif direction == self.RIGHT_DIRECTION or direction == self.TWISTED_RIGHT_DIRECTION then
    for z, actionBarRenderer in ipairs(self.actionBarRenderer) do
      actionBarRenderer:SetPosition(self.frame, 10+self.timerSkin.horizontalGap+self.actionBarSkin.horizontalGap/2,
                                    5+gap*(z-.5));
      if direction == self.TWISTED_RIGHT_DIRECTION then
        actionBarRenderer:SetAngle(-math.pi/8);      
      else
        actionBarRenderer:SetAngle(0);
      end
    end
    for z, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetPosition(self.frame, 5, 
                                      5+gap*(z-.5));
      visualTimerRenderer:SetDirection(2);  
    end
  else
    return;
  end
  local maxz = max(#self.actionBarRenderer, #self.visualTimerRenderer);
  self.frame:SetWidth(15+self.actionBarSkin.horizontalGap+self.timerSkin.horizontalGap);
  self.frame:SetHeight(10+maxz*gap);
  self.backup.direction = direction;
end

function ManagerFrame:GetDirection()
  return self.backup.direction;
end

function ManagerFrame:IsVisible()
  return self.backup.visible;
end

function ManagerFrame:ToggleVisibility()
  if self.backup.visible then
    self.frame:Hide();
  else
    self.frame:Show();      
  end    
end

--[[
  The ManagerFrameFactory class.
--]]

function ManagerFrameFactory:Create(frame, barNameList, actionBarConfMap, actionType, removeMacroTextMap, backup, messages, actionRendererCreatorFunc)
  managerFrame = self:NewInstance();
  local barNameListSize = #barNameList;
  if backup ~=nil and backup.left ~= nil then
    frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", backup.left, backup.top);
  end
  frame:SetWidth(270);
  frame:SetHeight(37*barNameListSize);
  frame:SetBackdrop(self.bckd);
  frame:SetBackdropColor(0 ,0 ,0 , 0);
  frame:SetBackdropBorderColor(.3, .3, .3, 0);
  frame:EnableMouse(false);
  frame:RegisterForDrag("LeftButton");
  frame:SetScript("OnDragStop",function (self)
    self:StopMovingOrSizing();
    self.managerFrame.backup.top = self:GetTop();
    self.managerFrame.backup.left = self:GetLeft(); 
  end);
  frame:SetScript("OnDragStart",function (self)
    self:StartMoving();
  end);  
  frame:Show();
  managerFrame.frame = frame;
  frame.managerFrame = managerFrame;
  managerFrame.backup = backup;
  
  managerFrame.actionBarRenderer = {};
  managerFrame.visualTimerRenderer = {};
  
  for z=1, barNameListSize do
    local barName = barNameList[z];
    local actionBarConfList = actionBarConfMap[barName];
    local actionList = {};
    local actionPropertyMap = {};
    local actionFuncMap = {};
    for i=1,#actionBarConfList do
      local actionBarConf = actionBarConfList[i];
      if actionBarConf.action==nil or actionBarConf.action=="spell" then
        local spellAction = actionLib.spellActionFactory:Create(actionBarConf.name, actionBarConf.target);
        actionPropertyMap[spellAction] = actionBarConf;
        if actionType == "totem" then
          if actionBarConf.totemName == nil then
            actionBarConf.totemName = actionBarConf.name;
          end
        elseif actionType == "buff" then
          if actionBarConf.buffName == nil then
            actionBarConf.buffName = actionBarConf.name;
          end        
        end
        if barName=="earthShield" then
          -- hook the timer duration (set it to 0 if you put earth shield on yourself)
          actionFuncMap[spellAction] = earthShieldTransformSpellArg;
        else
          actionFuncMap[spellAction] = transformSpellArg;
        end
        table.insert(actionList, spellAction);
      elseif actionBarConf.action=="item" then
        local itemAction = actionLib.itemActionFactory:Create(actionBarConf.name, actionBarConf.target, actionBarConf.equipItem);
        actionPropertyMap[itemAction] = actionBarConf;
        actionFuncMap[itemAction] = transformItemArg;
        table.insert(actionList, itemAction);
      end
    end
        
    local actionBarRenderer = actionRendererCreatorFunc();
    table.insert(managerFrame.actionBarRenderer, actionBarRenderer);
    local actionBar = actionBarLib.actionBarFactory:Create(
       barName.."Selected", -- this arg is used to name caster button
       actionList,
       actionBarRenderer,
       totemManagerData.actionBarBackup[barName]
    );
    totemManagerData.actionBarBackup[barName] = actionBar:GetBackup();
    actionBar:SetSelectedCtrlClickMacroText(removeMacroTextMap[barName]);

    local visualTimerRenderer = ManagerFrame.timerSkin.visualRenderer:Create(barName, 1);
    table.insert(managerFrame.visualTimerRenderer, visualTimerRenderer);
    local rendererList = {visualTimerRenderer};
    local timerMessage = libSinkTimerRendererLib.libSinkTimerRendererFactory:Create(
      messages.started, messages.stopping, messages.stopped, messages.countAtZero
    );
    if actionType == "totem" then
       timerMessage:ActivatePour(true, true, true, true);
       table.insert(rendererList, timerMessage);       
    elseif actionType == "buff" then
       -- only expiring
       timerMessage:ActivatePour(false, true);
       table.insert(rendererList, timerMessage);
    end
    local timer = timerLib.actionPerformedTimerFactory:Create(
      rendererList,
      actionList,
      actionPropertyMap,
      actionFuncMap
    );
  end
  
  managerFrame:SetDirection(nil, backup.direction);
  if not backup.visible then
    managerFrame:ToggleVisibility();
  end
  
  return managerFrame;
end

local bckd = {};
bckd.bgFile = "Interface/Tooltips/UI-Tooltip-Background";
bckd.edgeFile = "Interface/Tooltips/UI-Tooltip-Border";
bckd.tile = true;
bckd.tileSize = 16;
bckd.edgeSize = 16;
bckd.insets = {};
bckd.insets.left = 4;
bckd.insets.right = 4;
bckd.insets.top = 4;
bckd.insets.bottom = 4;
ManagerFrameFactory.bckd = bckd;

